#!/usr/bin/env python3

# Copyright 2017 Patrick O. Perry.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

try:
    import property
    import unicode_data
except ModuleNotFoundError:
    from util import property
    from util import unicode_data


DERIVED_CORE_PROPERTIES = "data/ucd/DerivedCoreProperties.txt"
EAST_ASIAN_WIDTH = "data/ucd/EastAsianWidth.txt"
EMOJI_DATA = "data/ucd/emoji/emoji-data.txt"


east_asian_width = property.read(EAST_ASIAN_WIDTH)
# A:  Ambiguous (can be narrow or wide depending on context; treat as wide)
# F:  Fullwidth (always wide)
# H:  Halfwidth (always narrow)
# Na: Narrow    (always narrow)
# N:  Neutral   (non-East Asian; treat as single)
# W:  Wide      (always wide)


emoji_props = property.read(EMOJI_DATA, sets=True)

# https://www.unicode.org/reports/tr51/#def_basic_emoji_set
emoji = ((emoji_props['Emoji'] - emoji_props['Emoji_Component'])
         & emoji_props['Emoji_Presentation'])

# Treat ignorables as invisible
derived_core_properties = property.read(DERIVED_CORE_PROPERTIES, sets=True)
default_ignorable = derived_core_properties['Default_Ignorable_Code_Point']


# unassigned: not assigned, other, surrogate
none_cats = set(['Cc', 'Cn', 'Co', 'Cs', 'Zl', 'Zp'])
mark_cats = set(['Cf', 'Me', 'Mn'])
none = set([0xFFF9, 0xFFFA, 0xFFFB]) # interlinear annotation markers
mark = set()
for code in range(len(unicode_data.uchars)):
    u = unicode_data.uchars[code]
    if code in none or code in mark:
        pass
    elif u is None or u.category in none_cats:
        none.add(code)
    elif u.category in mark_cats:
        mark.add(code)


code_props = [None] * len(east_asian_width)
for code in range(len(code_props)):
    eaw = east_asian_width[code]
    if code in default_ignorable: # default ingorable overrides
        code_props[code] = 'Ignorable'
    elif code in emoji: # emoji overrides
        code_props[code] = 'Emoji'
    elif code in mark: # mark overrides
        code_props[code] = 'Mark'
    elif code in none: # none overrides
        code_props[code] = 'None'
    elif eaw == 'F' or eaw == 'W':
        code_props[code] = 'Wide'
    elif eaw == 'H' or eaw == 'Na' or eaw == 'N':
        code_props[code] = 'Narrow'
    elif eaw == 'A':
        code_props[code] = 'Ambiguous'
    else:
        code_props[code] = 'Narrow' # default to narrow


prop_names = [
        'None', 'Ignorable', 'Mark', 'Narrow', 'Ambiguous', 'Wide', 'Emoji'
        ]
prop_vals = {}
for p in prop_names:
    prop_vals[p] = len(prop_vals)


def compute_tables(block_size):
    nblock = len(code_props) // block_size
    stage1 = [None] * nblock
    stage2 = []
    stage2_dict = {}
    for i in range(nblock):
        begin = i * block_size
        end = begin + block_size
        block = tuple(code_props[begin:end])
        if block in stage2_dict:
            j = stage2_dict[block]
        else:
            j = len(stage2)
            stage2_dict[block] = j
            stage2.append(block)
        stage1[i] = j
    return (stage1,stage2)


def stage1_item_size(nstage2):
    nbyte = math.ceil(math.log(nstage2, 2) / 8)
    size = 2**math.ceil(math.log(nbyte, 2))
    return size


page_size = 4096
block_size = 256

nbytes = {}

best_block_size = 1
smallest_size = len(code_props)

for i in range(1,17):
    block_size = 2**i
    stage1,stage2 = compute_tables(block_size)

    nbyte1 = len(stage1) * stage1_item_size(len(stage2))
    nbyte2 = len(stage2) * block_size

    nbyte1 = math.ceil(nbyte1 / page_size) * page_size
    nbyte2 = math.ceil(nbyte2 / page_size) * page_size
    nbyte = nbyte1 + nbyte2
    nbytes[block_size] = nbyte

    if nbyte < smallest_size:
        smallest_size = nbyte
        best_block_size = block_size


block_size = best_block_size
stage1,stage2 = compute_tables(block_size)

type1_size = stage1_item_size(len(stage2))

if type1_size == 1:
    type1 = 'uint8_t'
elif type1_size == 2:
    type1 = 'uint16_t'
elif type1_size == 4:
    type1 = 'uint32_t'
else:
    type1 = 'uint64_t'

type2 = 'int8_t'


# Write chardwidth.h to stdout

print("/* This file is automatically generated. DO NOT EDIT!")
print("   Instead, edit gen-charwidth.py and re-run.  */")
print("")
print("/*")
print(" * Unicode East_Asian_Width property values.")
print(" *")
print(" * Defined in UAX #11 \"East Asian Width\"")
print(" *")
print(" *     http://www.unicode.org/reports/tr11/")
print(" *")
print(" * We use the two-stage lookup strategy described at")
print(" *")
print(" *     http://www.strchr.com/multi-stage_tables")
print(" *")
print(" */")
print("")
print("#ifndef CHARWIDTH_H")
print("#define CHARWIDTH_H")
print("")
print("#include <stdint.h>")
print("")
print("enum charwidth_prop {")
first = True
for prop in prop_names:
    if not first:
        print(",\n", end="")
    else:
        first = False
    print("\tCHARWIDTH_" + prop.upper() + " = " + str(prop_vals[prop]), end="")
print("\n};")
print("")
print("static const " + type1 + " charwidth_stage1[] = {")
for i in range(len(stage1) - 1):
    if i % 16  == 0:
        print("/* U+{:04X} */".format(i * block_size), end="")
    print("{0: >3},".format(stage1[i]), end="")
    if i % 16 == 15:
        print("")
print("{0: >3}".format(stage1[len(stage1) - 1]))
print("};")
print("")
print("static const " + type2 + " charwidth_stage2[][" +
        str(block_size) + "] = {")
for i in range(len(stage2)):
    print("  /* block " + str(i) + " */")
    print("  {", end="")
    for j in range(block_size):
        print("{0: >3}".format(prop_vals[stage2[i][j]]), end="")
        if j + 1 == block_size:
            print("\n  }", end="")
        else:
            print(",", end="")
            if j % 16 == 15:
                print("\n   ", end="")
    if i + 1 != len(stage2):
        print(",\n")
    else:
        print("")
print("};")

print("")
print("static int charwidth(int32_t code)")
print("{")
print("\tconst int32_t block_size = " + str(block_size) + ";")
print("\t" + type1 + " i = charwidth_stage1[code / block_size];")
print("\treturn charwidth_stage2[i][code % block_size];")
print("}")
print("")
print("#endif /* CHARWIDTH_H */")



#!/usr/bin/env python3

import math
import re

CASE_FOLDING = 'data/ucd/CaseFolding.txt'
UNICODE_MAX = 0x10FFFF

# Parse CaseFolding.txt

try:
    file = open(CASE_FOLDING, 'r')
except FileNotFoundError:
    file = open('../' + CASE_FOLDING, 'r')

def utf8_len(code):
    if code <= 0x7f:
        return 1
    elif code <= 0x07FF:
        return 2
    elif code <= 0xFFFF:
        return 3
    else:
        return 4

with file:
    for line in file:
        if line[0] != '#' and line[0] != '\n':
            fields = line.split(';')
            code = int(fields[0], 16)
            status = fields[1].strip();

            if status == 'C' or status == 'F':
                l0 = utf8_len(code)

                mapping = [int(x,16) for x in fields[2].split()]
                l1 = sum([utf8_len(m) for m in mapping])

                ratio = l1 / l0
                if ratio >= 3:
                    print('U+{:04X}'.format(code), mapping, 'ratio: ', ratio)




#!/usr/bin/env python3

# Copyright 2017 Patrick O. Perry.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re

PATTERN = re.compile(r"""^([0-9A-Fa-f]+)        # (first code)
                          (\.\.([0-9A-Fa-f]+))? # (.. last code)?
                          \s*
                          ;                     # ;
                          \s*
                          (\w+)                 # (property name)
                          \s*
                          ;                     # ;
                          \s*
                          (\w+)                 # (property value)
                          \s*
                          (\#.*)?$              # (# comment)?""", re.X)

UNICODE_MAX = 0x10FFFF

DERIVED_NORMALIZATION_PROPS = "data/ucd/DerivedNormalizationProps.txt"
try:
    infile = open(DERIVED_NORMALIZATION_PROPS)
except FileNotFoundError:
    infile = open("../" + DERIVED_NORMALIZATION_PROPS, "r")

code_props = ['Yes'] * (UNICODE_MAX + 1)

for line in infile:
    m = PATTERN.match(line)
    if m is None or m.group(4) != 'NFC_QC':
        continue
    begin = int(m.group(1), 16)
    if m.group(3) is None:
        end = begin + 1
    else:
        end = int(m.group(3), 16) + 1
    prop = m.group(5)
    if prop == 'M':
        prop = 'Maybe'
    elif prop == 'N':
        prop = 'No'
    elif prop == 'Y':
        prop = 'Yes'
    for code in range(begin, end):
        code_props[code] = prop
    #print(line, end = "")
    #print('[', '{:04X}'.format(begin), ',', '{:04X}'.format(end), '): ',
    #      prop, sep = '')
infile.close()


prop_names = set(code_props)

prop_vals = {}
prop_vals['No'] = 0
prop_vals['Yes'] = 1
prop_vals['Maybe'] = -1

def compute_tables(block_size):
    nblock = len(code_props) // block_size
    stage1 = [None] * nblock
    stage2 = []
    stage2_dict = {}
    for i in range(nblock):
        begin = i * block_size
        end = begin + block_size
        block = tuple(code_props[begin:end])
        if block in stage2_dict:
            j = stage2_dict[block]
        else:
            j = len(stage2)
            stage2_dict[block] = j
            stage2.append(block)
        stage1[i] = j
    return (stage1,stage2)

def stage1_item_size(nstage2):
    nbyte = math.ceil(math.log(nstage2, 2) / 8)
    size = 2**math.ceil(math.log(nbyte, 2))
    return size

page_size = 4096
block_size = 256

nbytes = {}

best_block_size = 1
smallest_size = len(code_props)

for i in range(1,17):
    block_size = 2**i
    stage1,stage2 = compute_tables(block_size)
    #
    nbyte1 = len(stage1) * stage1_item_size(len(stage2))
    nbyte2 = len(stage2) * block_size
    #
    nbyte1 = math.ceil(nbyte1 / page_size) * page_size
    nbyte2 = math.ceil(nbyte2 / page_size) * page_size
    nbyte = nbyte1 + nbyte2
    nbytes[block_size] = nbyte
    #
    if nbyte < smallest_size:
        smallest_size = nbyte
        best_block_size = block_size


block_size = best_block_size
stage1,stage2 = compute_tables(block_size)

type1_size = stage1_item_size(len(stage2))

if type1_size == 1:
    type1 = 'uint8_t'
elif type1_size == 2:
    type1 = 'uint16_t'
elif type1_size == 4:
    type1 = 'uint32_t'
else:
    type1 = 'uint64_t'

type2 = 'int8_t'



# Write normalizationprop.h to stdout

print("/* This file is automatically generated. DO NOT EDIT!")
print("   Instead, edit gen-normalization.py and re-run.  */")
print("")
print("/*")
print(" * Unicode NFC_QC property values.")
print(" *")
print(" * Defined in UAX #15 \"Unicode Normalization Forms\"")
print(" *")
print(" *     http://www.unicode.org/reports/tr15/")
print(" *")
print(" * Section 9, \"Detecting Normalization Forms.\"")
print(" *")
print(" *")
print(" * We use the two-stage lookup strategy described at")
print(" *")
print(" *     http://www.strchr.com/multi-stage_tables")
print(" *")
print(" */")
print("")
print("#ifndef NORMALIZATIONPROP_H")
print("#define NORMALIZATIONPROP_H")
print("")
print("#include <stdint.h>")
print("")
print("enum nfc_qc_prop {")
for i in range(len(prop_names)):
    prop = sorted(prop_names)[i]
    if i > 0:
        print(",")
    print("\tNFC_QC_" + prop.upper() + " = " + str(prop_vals[prop]), end="")
print("\n};")
print("")
print("static const " + type1 + " nfc_qc_stage1[] = {")
for i in range(len(stage1) - 1):
    if i % 16  == 0:
        print("/* U+{:04X} */".format(i * block_size), end="")
    print("{0: >3},".format(stage1[i]), end="")
    if i % 16 == 15:
        print("")
print("{0: >3}".format(stage1[len(stage1) - 1]))
print("};")
print("")
print("static const " + type2 + " nfc_qc_stage2[][" +
        str(block_size) + "] = {")
for i in range(len(stage2)):
    print("  /* block " + str(i) + " */")
    print("  {", end="")
    for j in range(block_size):
        print("{0: >3}".format(prop_vals[stage2[i][j]]), end="")
        if j + 1 == block_size:
            print("\n  }", end="")
        else:
            print(",", end="")
            if j % 16 == 15:
                print("\n   ", end="")
    if i + 1 != len(stage2):
        print(",\n")
    else:
        print("")
print("};")

print("")
print("static int nfc_qc(uint32_t code)")
print("{")
print("\tconst uint32_t block_size = " + str(block_size) + ";")
print("\t" + type1 + " i = nfc_qc_stage1[code / block_size];")
print("\treturn nfc_qc_stage2[i][code % block_size];")
print("}")
print("")
print("#endif /* NORMALIZATIONPROP_H */")



# unicode_data

import collections
import re


UNICODE_DATA = 'data/ucd/UnicodeData.txt'
UNICODE_MAX = 0x10FFFF

try:
    unicode_data = open(UNICODE_DATA, "r")
except FileNotFoundError:
    unicode_data = open("../" + UNICODE_DATA, "r")

field_names = [
    'name',     # Name
    'category', # General_Category
    'ccc',      # Canonical_Combining_Class
    'bidi',     # Bidi_Class
    'decomp',   # Decomposition_Type, Decomposition_Mapping
    'decimal',  # Numeric_Value (Numeric_Type = Decimal)
    'digit',    # Numeric_Value (Numeric_Type = Decimal, Digit)
    'numeric',  # Numeric_Value (Numeric_Type = Decimal, Digit, Numeric)
    'mirrored', # Bidi_Mirrored
    'old_name', # Unicode_1_Name
    'comment',  # ISO_Comment
    'ucase',    # Simple_Uppercase_Mapping
    'lcase',    # Simple_Lowercase_Mapping
    'tcase'     # Simple_Titlecase_Mapping
    ]

UChar = collections.namedtuple('UChar', field_names)
ids = UChar._make(range(1, len(field_names) + 1))

Decomp = collections.namedtuple('Decomp', ['type', 'map'])

DECOMP_PATTERN = re.compile(r"""^(<(\w+)>)?\s* # decomposition type
                                 ((\s*[0-9A-Fa-f]+)+) # decomposition mapping
                                 \s*$""", re.X)
RANGE_PATTERN = re.compile(r"""^<([^,]+),\s*    # range name
                                 (First|Last) # first or last
                                 >$""", re.X)

def parse_decomp(code, field):
    if field != '':
        m = DECOMP_PATTERN.match(field)
        assert m
        d_type = m.group(2)
        d_map = tuple([int(x, 16) for x in m.group(3).split()])
        return Decomp(type=d_type, map=d_map)

    elif code in range(0xAC00, 0xD7A4):
        return Decomp(type='hangul', map=None)
    else:
        return None


def parse_code(field):
    if field != '':
        return int(field, 16)
    else:
        return None

def parse_int(field):
    if field != '':
        return int(field)
    else:
        return None

def parse_str(field):
    if field == '':
        return None
    else:
        return field


uchars = [None] * (UNICODE_MAX + 1)

with unicode_data:
    for line in unicode_data:
        fields = line.strip().split(';')
        code = int(fields[0], 16)
        uchars[code] = UChar(name = fields[ids.name],
                             category = parse_str(fields[ids.category]),
                             ccc = parse_int(fields[ids.ccc]),
                             bidi = fields[ids.bidi],
                             decomp = parse_decomp(code, fields[ids.decomp]),
                             decimal = fields[ids.decimal],
                             digit = fields[ids.digit],
                             numeric = fields[ids.numeric],
                             mirrored = fields[ids.mirrored],
                             old_name = fields[ids.old_name],
                             comment = fields[ids.comment],
                             ucase = parse_code(fields[ids.ucase]),
                             lcase = parse_code(fields[ids.lcase]),
                             tcase = parse_code(fields[ids.tcase]))


utype = None

for code in range(len(uchars)):
    u = uchars[code]
    if u is None:
        uchars[code] = utype
    else:
        m = RANGE_PATTERN.match(u.name)
        if m:
            if m.group(2) == 'First':
                utype = u._replace(name = '<' + m.group(1) + '>')
            else:
                utype = None



decomp_vals = {
    'hangul': -1, 'none': 0,
    'font': 1, 'noBreak': 2, 'initial': 3, 'medial': 4, 'final': 5,
    'isolated': 6, 'circle': 7, 'super': 8, 'sub': 9, 'vertical': 10,
    'wide': 11, 'narrow': 12, 'small': 13, 'square': 14, 'fraction': 15,
    'compat': 16 }

decomp_map = []
decomp = []

for code in range(len(uchars)):
    u = uchars[code]

    if u is None or u.decomp is None:
        decomp.append(None)
        continue

    d = u.decomp
    if d.map is not None:
        d_len = len(d.map)

        if d_len > 1:
            d_data = len(decomp_map)
            decomp_map.extend(d.map)
        else:
            d_data = d.map[0]

        decomp.append((d.type, d_len, d_data))

    elif d.type == 'hangul':
        decomp.append(('hangul', 2, 0))

    else:
        decomp.append(None)


# From Unicode-8.0 Section 3.12 Conjoining Jamo Behavior
HANGUL_SBASE = 0xAC00
HANGUL_LBASE = 0x1100
HANGUL_VBASE = 0x1161
HANGUL_TBASE = 0x11A7
HANGUL_LCOUNT = 19
HANGUL_VCOUNT = 21
HANGUL_TCOUNT = 28
HANGUL_NCOUNT = (HANGUL_VCOUNT * HANGUL_TCOUNT)
HANTUL_SCOUNT = (HANGUL_LCOUNT * HANGUL_NCOUNT)


def hangul_decompose(code):
    sindex = code - HANGUL_SBASE
    lindex = sindex // HANGUL_NCOUNT
    vindex = (sindex % HANGUL_NCOUNT) // HANGUL_TCOUNT
    tindex = sindex % HANGUL_TCOUNT
    lpart = HANGUL_LBASE + lindex
    vpart = HANGUL_VBASE + vindex;
    tpart = HANGUL_TBASE + tindex;
    if tindex > 0:
        return (lpart, vpart, tpart)
    else:
        return (lpart, vpart)


# get a character's decomposition if one exists, as a tuple
def decompose(code, compat=True):
    dc = decomp[code]
    if dc is None:
        return None
    t = dc[0]
    l = dc[1]
    if not compat and t is not None and t != 'hangul':
        return None
    if l == 1:
        return (dc[2],)
    elif dc[0] != 'hangul':
        o = dc[2]
        return tuple(decomp_map[o:o + l])
    else:
        return hangul_decompose(code)



#!/usr/bin/env python3

# Copyright 2017 Patrick O. Perry.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re

try:
    import unicode_data
except ModuleNotFoundError:
    from util import unicode_data


EXCLUSIONS = 'data/ucd/CompositionExclusions.txt'

# get the length-2 decomposition maps (excluding hangul and compatibility maps)

decomp_map = {}
starter = [None] * len(unicode_data.uchars)

for code in range(len(unicode_data.uchars)):
    u = unicode_data.uchars[code]
    if u is None:
        continue

    ccc = u.ccc
    if ccc is None or ccc == 0:
        starter[code] = True
    else:
        starter[code] = False

    d = u.decomp
    if d is not None and d.type is None:
        if len(d.map) == 2:
            decomp_map[code] = tuple(d.map)


# exclude non-starter decomposiitons

decomp_map2 = {}
for p,d in decomp_map.items():
    if starter[p] and starter[d[0]]:
        decomp_map2[p] = d
decomp_map = decomp_map2


# exclude composition exclusions

try:
    file = open(EXCLUSIONS, 'r')
except FileNotFoundError:
    file = open('../' + EXCLUSIONS, 'r')

with file:
    for line in file:
        fields = line.partition('#')
        code = fields[0].strip()
        if len(code) > 0:
            code = int(code, 16)
            if code in decomp_map:
                del decomp_map[code]

#print('primary\tletter\tcode')
#for p,d in decomp_map.items():
#    print(p, '\t', d[0], '\t', d[1], sep='')

# construct table l : [(c,p)]
compose_map = {}
for p,d in decomp_map.items():
    l = d[0]
    c = d[1]
    if l not in compose_map:
        compose_map[l] = []
    compose_map[l].append((c, p))

compose = []
combiner = []
primary = []
off = 0
for code in range(len(unicode_data.uchars)):
    if code in compose_map:
        maps = compose_map[code]
        maps.sort()
        compose.append((off, len(maps)))
        combiner.extend([c for (c,p) in maps])
        primary.extend([p for (c,p) in maps])
        off += len(maps)
    else:
        compose.append((0,0))

# Hangul
hangul_lpart = off
hangul_lvpart = off + 1

for code in range(0x1100, 0x1113):
    compose[code] = (hangul_lpart, 1)

for code in range(0xAC00, 0xD7A4):
    if (code - 0xAC00) % 28 == 0:
        compose[code] = (hangul_lvpart, 1)


def compute_tables(block_size):
    nblock = len(compose) // block_size
    stage1 = [None] * nblock
    stage2 = []
    stage2_dict = {}
    for i in range(nblock):
        begin = i * block_size
        end = begin + block_size
        block = tuple(compose[begin:end])
        if block in stage2_dict:
            j = stage2_dict[block]
        else:
            j = len(stage2)
            stage2_dict[block] = j
            stage2.append(block)
        stage1[i] = j
    return (stage1,stage2)


def stage1_item_size(nstage2):
    nbyte = math.ceil(math.log(nstage2, 2) / 8)
    size = 2**math.ceil(math.log(nbyte, 2))
    return size

page_size = 4096
block_size = 256

nbytes = {}

best_block_size = 1
smallest_size = len(compose)

for i in range(1,17):
    block_size = 2**i
    stage1,stage2 = compute_tables(block_size)

    nbyte1 = len(stage1) * stage1_item_size(len(stage2))
    nbyte2 = len(stage2) * block_size

    nbyte1 = math.ceil(nbyte1 / page_size) * page_size
    nbyte2 = math.ceil(nbyte2 / page_size) * page_size
    nbyte = nbyte1 + nbyte2
    nbytes[block_size] = nbyte

    if nbyte < smallest_size:
        smallest_size = nbyte
        best_block_size = block_size


block_size = best_block_size
stage1,stage2 = compute_tables(block_size)

type1_size = stage1_item_size(len(stage2))
if type1_size == 1:
    type1 = 'uint8_t'
elif type1_size == 2:
    type1 = 'uint16_t'
elif type1_size == 4:
    type1 = 'uint32_t'
else:
    type1 = 'uint64_t'

type2 = 'struct composition'


# Write compose.h to stdout


print("/* This file is automatically generated. DO NOT EDIT!")
print("   Instead, edit gen-compose.py and re-run. */")
print("")
print("/*")
print(" * Unicode primary composites.")
print(" *")
print(" * Defined in Unicode Sec 3.11 \"Normalization Forms\"")
print(" *")
print(" * We use the two-stage lookup strategy described at")
print(" *")
print(" *     http://www.strchr.com/multi-stage_tables")
print(" *")
print(" */")
print("")
print("#ifndef UNICODE_COMPOSE_H")
print("#define UNICODE_COMPOSE_H")
print("")
print("#include <stdint.h>")
print("")
print("/* composition")
print(" * -----------")
print(" * offset: the offset into the primary and combiner arrays,")
print(" *         or 0 if there are no compositions")
print(" * length: the number of compositions for the codepont")
print(" */")
print("struct composition {")
print("\tunsigned offset : 11;")
print("\tunsigned length : 5;")
print("};")
print("")
print("#define COMPOSITION_BLOCK_SIZE", block_size)
print("")
print("#define COMPOSITION_HANGUL_LPART", hangul_lpart)
print("")
print("#define COMPOSITION_HANGUL_LVPART", hangul_lvpart)
print("")
print("static const " + type1 + " composition_stage1[] = {")
for i in range(len(stage1) - 1):
    if i % 16  == 0:
        print("/* U+{:04X} */".format(i * block_size), end="")
    print("{0: >3},".format(stage1[i]), end="")
    if i % 16 == 15:
        print("")
print("{0: >3}".format(stage1[len(stage1) - 1]))
print("};")
print("")
print("static const " + type2 + " composition_stage2[][" +
        str(block_size) + "] = {")
for i in range(len(stage2)):
    print("  /* block " + str(i) + " */")
    print("  {", end="")
    for j in range(block_size):
        print("{{{0: >3}".format(stage2[i][j][0]), end="")
        print(",{0: >2}}}".format(stage2[i][j][1]), end="")
        if j + 1 == block_size:
            print("\n  }", end="")
        else:
            print(",", end="")
            if j % 7 == 6:
                print("\n   ", end="")
            else:
                print(" ", end="")
    if i + 1 != len(stage2):
        print(",\n")
    else:
        print("")
print("};")
print("")
print("static const int32_t composition_combiner[] = {")
for i in range(len(combiner) - 1):
    if i % 8  == 0:
        print("/* {0: >3} */ ".format(i), end="")
    print("0x{0:04X},".format(combiner[i]), end="")
    if i % 8 == 7:
        print("")
print("0x{0:04X}".format(combiner[len(combiner) - 1]))
print("};")
print("")
print("static const int32_t composition_primary[] = {")
for i in range(len(primary) - 1):
    if i % 8  == 0:
        print("/* {0: >3} */ ".format(i), end="")
    print("0x{0:04X},".format(primary[i]), end="")
    if i % 8 == 7:
        print("")
print("0x{0:04X}".format(primary[len(primary) - 1]))
print("};")
print("")
print("#endif /* UNICODE_COMPOSE_H */")



#!/usr/bin/env python3

# Copyright 2017 Patrick O. Perry.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
import math

try:
    import unicode_data
except ModuleNotFoundError:
    from util import unicode_data

decomp_vals = unicode_data.decomp_vals
decomp_map = unicode_data.decomp_map
decomp = unicode_data.decomp

def compute_tables(block_size):
    nblock = len(decomp) // block_size
    stage1 = [None] * nblock
    stage2 = []
    stage2_dict = {}
    for i in range(nblock):
        begin = i * block_size
        end = begin + block_size
        block = tuple(decomp[begin:end])
        if block in stage2_dict:
            j = stage2_dict[block]
        else:
            j = len(stage2)
            stage2_dict[block] = j
            stage2.append(block)
        stage1[i] = j
    return (stage1,stage2)


def stage1_item_size(nstage2):
    nbyte = math.ceil(math.log(nstage2, 2) / 8)
    size = 2**math.ceil(math.log(nbyte, 2))
    return size

page_size = 4096
block_size = 256

nbytes = {}

best_block_size = 1
smallest_size = len(decomp)

for i in range(1,17):
    block_size = 2**i
    stage1,stage2 = compute_tables(block_size)

    nbyte1 = len(stage1) * stage1_item_size(len(stage2))
    nbyte2 = len(stage2) * block_size

    nbyte1 = math.ceil(nbyte1 / page_size) * page_size
    nbyte2 = math.ceil(nbyte2 / page_size) * page_size
    nbyte = nbyte1 + nbyte2
    nbytes[block_size] = nbyte

    if nbyte < smallest_size:
        smallest_size = nbyte
        best_block_size = block_size


block_size = best_block_size
stage1,stage2 = compute_tables(block_size)

type1_size = stage1_item_size(len(stage2))
if type1_size == 1:
    type1 = 'uint8_t'
elif type1_size == 2:
    type1 = 'uint16_t'
elif type1_size == 4:
    type1 = 'uint32_t'
else:
    type1 = 'uint64_t'

# Write decompose.h to stdout

print("/* This file is automatically generated. DO NOT EDIT!")
print("   Instead, edit gen-decompose.py and re-run.  */")
print("")
print("/*")
print(" * Decomposition mappings.")
print(" *")
print(" * Defined in UAX #44 \"Unicode Character Database\"")
print(" *")
print(" *     http://www.unicode.org/reports/tr44/")
print(" *")
print(" * Section 5.7.3, Table 14.")
print(" *")
print(" *")
print(" * We use a two-stage lookup strategy as described at")
print(" *")
print(" *     http://www.strchr.com/multi-stage_tables")
print(" *")
print(" */")
print("")
print("#ifndef UNICODE_DECOMPOSE_H")
print("#define UNICODE_DECOMPOSE_H")
print("")
print("#include <stdint.h>")
print("")
print("/* decomposition_type")
print(" * ------------------")
print(" * compatibility decompositions have decomposition_type != 0")
print(" */")
print("enum decomposition_type {", end="")
first = True
for k,v in sorted(decomp_vals.items(), key=operator.itemgetter(1)):
    if not first:
        print(",", end="")
    print("\n\tDECOMPOSITION_" + k.upper() + " = " + str(v), end="")
    first = False
print("\n};")
print("")
print("/* decomposition")
print(" * -------------")
print(" * type:   the decomposition_type")
print(" *")
print(" * length: the length (in codepoints) of the decomposition mapping,")
print(" *         or 0 if there is no decomposition")
print(" *")
print(" * data:   the mapped-to codepoint (length = 1), or")
print(" *         an index into the `decomposition_mapping` array, pointing")
print(" *         to the first codepoint in the mapping (length > 1)")
print(" */")
print("struct decomposition {")
print("\tint type : 6;")
print("\tunsigned length : 5;")
print("\tunsigned data : 21;")
print("};")
print("")
print("#define DECOMPOSITION_BLOCK_SIZE", block_size)
print("")
print("static const " + type1 + " decomposition_stage1[] = {")
for i in range(len(stage1) - 1):
    if i % 16  == 0:
        print("/* U+{:04X} */".format(i * block_size), end="")
    print("{0: >3},".format(stage1[i]), end="")
    if i % 16 == 15:
        print("")
print("{0: >3}".format(stage1[len(stage1) - 1]))
print("};")
print("")
print("static const struct decomposition decomposition_stage2[][" +
        str(block_size) + "] = {")
for i in range(0,len(stage2)):
    print("  /* block " + str(i) + " */")
    print("  {", end="")
    for j in range(block_size):
        val = stage2[i][j]
        if val is None:
            print("{0,0,0}", end="")
        else:
            if val[0] is None:
                t = 0
            else:
                t = decomp_vals[val[0]]
            print("{{{0},{1},0x{2:05X}}}".format(t, val[1], val[2]), end="")

        #print("{0: >3}".format(prop_vals[stage2[i][j]]), end="")
        if j + 1 == block_size:
            print("\n  }", end="")
        else:
            print(",", end="")
            if j % 5 == 4:
                print("\n   ", end="")
    if i + 1 != len(stage2):
        print(",\n")
    else:
        print("")
print("};")
print("")
print("static const int32_t decomposition_mapping[] = {")
for i in range(len(decomp_map) - 1):
    if i % 8  == 0:
        print("/* 0x{:04X} */ ".format(i), end="")
    print("0x{0:04X},".format(decomp_map[i]), end="")
    if i % 8 == 7:
        print("")
print("0x{0:04X}".format(decomp_map[len(decomp_map) - 1]))
print("};")
print("")
print("#endif /* UNICODE_DECOMPOSE_H */")




import re

PATTERN = re.compile(r"""^([0-9A-Fa-f]+)        # (first code)
                          (\.\.([0-9A-Fa-f]+))? # (.. last code)?
                          \s*
                          ;                     # ;
                          \s*
                          (\w+)                 # (property name)
                          \s*
                          (\#.*)?$              # (# comment)?""", re.X)

UNICODE_MAX = 0x10FFFF


def read(filename, sets=False):
    try:
        file = open(filename, "r")
    except FileNotFoundError:
        file = open("../" + filename, "r")
    
    code_props = [None] * (UNICODE_MAX + 1)
    prop_names = set()
    properties = {}

    with file:
        for line in file:
            line = line.split("#")[0] # remove comment
            m = PATTERN.match(line)
            if m:
                first = int(m.group(1), 16)
                if m.group(3):
                    last = int(m.group(3), 16)
                else:
                    last = first
                name = m.group(4)
                if not name in properties:
                    properties[name] = set()
                prop = properties[name]
                for u in range(first, last + 1):
                    if not sets:
                        assert code_props[u] is None
                    code_props[u] = name
                    prop.add(u)
                prop_names.add(name)
    if sets:
        return properties
    else:
        return code_props



#!/usr/bin/env python3

# Copyright 2017 Patrick O. Perry.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

try:
    import property
    import unicode_data
except ModuleNotFoundError:
    from util import property
    from util import unicode_data

EMOJI_DATA = "data/ucd/emoji/emoji-data.txt"

emoji_props = property.read(EMOJI_DATA, sets=True)


props = [0] * len(unicode_data.uchars)
for i, (key, values) in enumerate(emoji_props.items()):
    bit = 0x1 << i
    for value in values:
        props[value] |= bit


def compute_tables(block_size):
    nblock = len(props) // block_size
    stage1 = [None] * nblock
    stage2 = []
    stage2_dict = {}
    for i in range(nblock):
        begin = i * block_size
        end = begin + block_size
        block = tuple(props[begin:end])
        if block in stage2_dict:
            j = stage2_dict[block]
        else:
            j = len(stage2)
            stage2_dict[block] = j
            stage2.append(block)
        stage1[i] = j
    return (stage1,stage2)


def stage1_item_size(nstage2):
    nbyte = math.ceil(math.log(nstage2, 2) / 8)
    size = 2**math.ceil(math.log(nbyte, 2))
    return size

page_size = 4096
block_size = 256

nbytes = {}

best_block_size = 1
smallest_size = len(props)

for i in range(1,17):
    block_size = 2**i
    stage1,stage2 = compute_tables(block_size)

    nbyte1 = len(stage1) * stage1_item_size(len(stage2))
    nbyte2 = len(stage2) * block_size

    nbyte1 = math.ceil(nbyte1 / page_size) * page_size
    nbyte2 = math.ceil(nbyte2 / page_size) * page_size
    nbyte = nbyte1 + nbyte2
    nbytes[block_size] = nbyte

    if nbyte < smallest_size:
        smallest_size = nbyte
        best_block_size = block_size


block_size = best_block_size
stage1, stage2 = compute_tables(block_size)

type1_size = stage1_item_size(len(stage2))
if type1_size == 1:
    type1 = 'uint8_t'
elif type1_size == 2:
    type1 = 'uint16_t'
elif type1_size == 4:
    type1 = 'uint32_t'
else:
    type1 = 'uint64_t'

assert len(emoji_props) <= 8
type2 = 'uint8_t'

print("/* This file is automatically generated. DO NOT EDIT!")
print("   Instead, edit gen-emoji.py and re-run.  */")
print("")
print("/*")
print(" * Unicode Emoji property values.")
print(" *")
print(" * We use the two-stage lookup strategy described at")
print(" *")
print(" *     http://www.strchr.com/multi-stage_tables")
print(" *")
print(" */")
print("")
print("#ifndef UNICODE_EMOJIPROP_H")
print("#define UNICODE_EMOJIPROP_H")
print("")
print("#include <stdint.h>")
print("")
print("enum emoji_prop_type {")
print("\tEMOJI_PROP_NONE = 0", end="")
for i, name in enumerate(emoji_props.keys()):
    print(",\n\tEMOJI_PROP_" + name.upper() + " = (1 << " + str(i) + ")",
          end="")
print("\n};")
print("")
print("static const " + type1 + " emoji_prop_stage1[] = {")
for i in range(len(stage1) - 1):
    if i % 16  == 0:
        print("/* U+{:04X} */".format(i * block_size), end="")
    print("{0: >3},".format(stage1[i]), end="")
    if i % 16 == 15:
        print("")
print("{0: >3}".format(stage1[len(stage1) - 1]))
print("};")
print("")
print("static const " + type2 + " emoji_prop_stage2[][" +
        str(block_size) + "] = {")
for i in range(len(stage2)):
    print("  /* block " + str(i) + " */")
    print("  {", end="")
    for j in range(block_size):
        print("{0: >3}".format(stage2[i][j]), end="")
        if j + 1 == block_size:
            print("\n  }", end="")
        else:
            print(",", end="")
            if j % 16 == 15:
                print("\n   ", end="")
    if i + 1 != len(stage2):
        print(",\n")
    else:
        print("")
print("};")

print("")
print("static int emoji_prop(int32_t code)")
print("{")
print("\tconst int32_t block_size = " + str(block_size) + ";")
print("\t" + type1 + " i = emoji_prop_stage1[code / block_size];")
print("\treturn emoji_prop_stage2[i][code % block_size];")
print("}")
print("")
print("#endif /* UNICODE_EMOJIPROP_H */")



#!/usr/bin/env python3

# Copyright 2017 Patrick O. Perry.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re

CASE_FOLDING = 'data/ucd/CaseFolding.txt'
UNICODE_MAX = 0x10FFFF

# Parse CaseFolding.txt

try:
    file = open(CASE_FOLDING, 'r')
except FileNotFoundError:
    file = open('../' + CASE_FOLDING, 'r')

casefold = []
casefold_map = []

with file:
    for line in file:
        if line[0] != '#' and line[0] != '\n':
            fields = line.split(';')
            code = int(fields[0], 16)
            status = fields[1].strip();

            if status == 'C' or status == 'F':
                while code > len(casefold):
                    casefold.append(None)
                assert code == len(casefold)

                mapping = [int(x,16) for x in fields[2].split()]
                n = len(mapping)
                if n > 1:
                    casefold.append((n, len(casefold_map)))
                    casefold_map.extend(mapping)
                else:
                    casefold.append((1, mapping[0]))

while len(casefold) <= UNICODE_MAX:
    casefold.append(None)


def compute_tables(block_size):
    nblock = (UNICODE_MAX + 1) // block_size
    stage1 = [None] * nblock
    stage2 = []
    stage2_dict = {}
    for i in range(nblock):
        begin = i * block_size
        end = begin + block_size
        block = tuple(casefold[begin:end])
        if block in stage2_dict:
            j = stage2_dict[block]
        else:
            j = len(stage2)
            stage2_dict[block] = j
            stage2.append(block)
        stage1[i] = j
    return (stage1,stage2)


def stage1_item_size(nstage2):
    nbyte = max(1, math.ceil(math.log(nstage2, 2) / 8))
    size = 2**math.ceil(math.log(nbyte, 2))
    return size

page_size = 4096
nbytes = {}

best_block_size = 1
smallest_size = UNICODE_MAX + 1

for i in range(1,17):
    block_size = 2**i
    stage1,stage2 = compute_tables(block_size)

    nbyte1 = len(stage1) * stage1_item_size(len(stage2))
    nbyte2 = len(stage2) * block_size

    nbyte1 = math.ceil(nbyte1 / page_size) * page_size
    nbyte2 = math.ceil(nbyte2 / page_size) * page_size
    nbyte = nbyte1 + nbyte2
    nbytes[block_size] = nbyte

    if nbyte < smallest_size:
        smallest_size = nbyte
        best_block_size = block_size


block_size = best_block_size
stage1,stage2 = compute_tables(block_size)

type1_size = stage1_item_size(len(stage2))
if type1_size == 1:
    type1 = 'uint8_t'
elif type1_size == 2:
    type1 = 'uint16_t'
elif type1_size == 4:
    type1 = 'uint32_t'
else:
    type1 = 'uint64_t'

# Write casefold.h to stdout

print("/* This file is automatically generated. DO NOT EDIT!")
print("   Instead, edit gen-casefold.py and re-run.  */")
print("")
print("/*")
print(" * Case folding properties.")
print(" *")
print(" * Defined in UAX #44 \"Unicode Character Database\"")
print(" *")
print(" *     http://www.unicode.org/reports/tr44/")
print(" *")
print(" * Section 5.6, Case and Case Mapping")
print(" *")
print(" *")
print(" * We use a two-stage lookup strategy as described at")
print(" *")
print(" *     http://www.strchr.com/multi-stage_tables")
print(" *")
print(" */")
print("")
print("#ifndef UNICODE_CASEFOLD_H")
print("#define UNICODE_CASEFOLD_H")
print("")
print("#include <stdint.h>")
print("")
print("/* casefold")
print(" * --------")
print(" * length: the length (in codepoints) of the case fold mapping,")
print(" *         or 0 if there is no case fold")
print(" *")
print(" * data:   the mapped-to codepoint (length = 1), or")
print(" *         an index into the `casefold_mapping` array, pointing")
print(" *         to the first codepoint in the mapping (length > 1)")
print(" */")
print("struct casefold {")
print("\tunsigned length : 8;")
print("\tunsigned data : 24;")
print("};")
print("")
print("#define CASEFOLD_BLOCK_SIZE", block_size)
print("")
print("static const " + type1 + " casefold_stage1[] = {")
for i in range(len(stage1) - 1):
    if i % 16  == 0:
        print("/* U+{:04X} */".format(i * block_size), end="")
    print("{0: >3},".format(stage1[i]), end="")
    if i % 16 == 15:
        print("")
print("{0: >3}".format(stage1[len(stage1) - 1]))
print("};")
print("")
print("static const struct casefold casefold_stage2[][" +
        str(block_size) + "] = {")
for i in range(0,len(stage2)):
    print("  /* block " + str(i) + " */")
    print("  {", end="")
    for j in range(block_size):
        val = stage2[i][j]
        if val is None:
            print("{0,0}", end="")
        else:
            print("{{{0},0x{1:05X}}}".format(val[0], val[1]), end="")

        #print("{0: >3}".format(prop_vals[stage2[i][j]]), end="")
        if j + 1 == block_size:
            print("\n  }", end="")
        else:
            print(",", end="")
            if j % 5 == 4:
                print("\n   ", end="")
    if i + 1 != len(stage2):
        print(",\n")
    else:
        print("")
print("};")
print("")
print("static const int32_t casefold_mapping[] = {")
for i in range(len(casefold_map) - 1):
    if i % 8  == 0:
        print("/* 0x{:04X} */ ".format(i), end="")
    print("0x{0:04X},".format(casefold_map[i]), end="")
    if i % 8 == 7:
        print("")
print("0x{0:04X}".format(casefold_map[len(casefold_map) - 1]))
print("};")
print("")
print("#endif /* UNICODE_CASEFOLD_H */")
